# encoding: utf-8

import logging
from base64 import b64decode
from base64 import b64encode
from Crypto.Protocol.KDF import PBKDF2 as derive_key
from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes
from django.conf import settings
from .helpers import _as_bytes


__all__ = [
    u'encrypt',
    u'encrypt_bytes',
    u'decrypt',
    u'decrypt_bytes',
]


logger = logging.getLogger(__name__)


def _prep_key(key=None):
    key = key or settings.DJENGA_ENCRYPTION_KEY
    return _as_bytes(key)


def _gcm_pack(header, cipher_text, tag, nonce, kdf_salt):
    values = [ header, cipher_text, tag, nonce, kdf_salt ]
    values = [ b64encode(x).decode('utf-8') for x in values ]
    return '|'.join(values)


def _gcm_unpack(value):
    values = value.split('|')
    if len(values) != 5:
        raise ValueError('Unpacked value did not have exactly four pieces')
    values = [ b64decode(x.encode('utf-8')) for x in values ]
    return values


def encrypt(plain_text, key=None, auth_header='djenga'):
    """
    The encrypt function encrypts a unicode string using the
    Blowfish cipher (provided by pycrypto).  The key used is
    the SECRET_KEY specified in the settings file.

    :param plain_text: The plaintext unicode string to be encrypted.
    :param key: the password to use for encryption
    :param auth_header: str
    :return The encrypted ciphertext.

    """
    plain_text = _as_bytes(plain_text)
    return encrypt_bytes(plain_text, key, auth_header)


def encrypt_bytes(plain_text, key=None, auth_header='djenga'):
    """
    The encrypt function encrypts a unicode string using the
    Blowfish cipher (provided by pycrypto).  The key used is
    the SECRET_KEY specified in the settings file.

    :param plain_text: The plaintext unicode string to be encrypted.
    :param key: the password to use for encryption
    :param auth_header: str
    :return The encrypted ciphertext.

    """
    kdf_salt = get_random_bytes(32)
    key = key or settings.DJENGA_ENCRYPTION_KEY
    derived_key = derive_key(key, kdf_salt, 32)
    cipher = AES.new(derived_key, AES.MODE_GCM)
    auth_header = _as_bytes(auth_header)
    cipher.update(auth_header)
    cipher_text, tag = cipher.encrypt_and_digest(plain_text)
    nonce = cipher.nonce
    return _gcm_pack(auth_header, cipher_text, tag, nonce, kdf_salt)


def decrypt(packed_value, key=None):
    data = decrypt_bytes(packed_value, key)
    return data.decode('utf-8')


def decrypt_bytes(packed_value, key=None):
    """
    The decrypt function decrypts unicode strings that have
    been encrypted by the util.encryption.encrypt() function.
    The cipher used is Blowfish (provided by pcrypto), and the
    key used is the SECRET_KEY specified in the settings file.

    :param packed_value: The encrypted pieces needed for decryption.
    :param key: the password to use when encrypting
    :return The decrypted plaintext (unicode) string.

    >>> import uuid
    >>> key = uuid.uuid4()
    >>> st = 'hello'
    >>> decrypt(encrypt(st, key.bytes), key.bytes) == st
    True
    """
    header, cipher_text, tag, nonce, kdf_salt = _gcm_unpack(packed_value)
    key = key or settings.DJENGA_ENCRYPTION_KEY
    decryption_key = derive_key(key, kdf_salt, 32)
    cipher = AES.new(decryption_key, AES.MODE_GCM, nonce)
    cipher.update(header)
    data = cipher.decrypt_and_verify(cipher_text, tag)
    return data


def _test_me():
    import uuid
    plain_text = 'Hello sesame!'
    key = uuid.uuid4()
    cipher_text = encrypt(plain_text, key.bytes)
    print(f'cipher_text: {cipher_text}')
    value = decrypt(cipher_text, key.bytes)
    print(f'after      : {value}')


if __name__ == "__main__":
    _test_me()





from datetime import datetime
import json
import logging


__all__ = [
    'JsonFormatter',
    'JsonTaskFormatter',
]


class JsonFormatter(logging.Formatter):
    """
    This formatter is useful if you want to ship logs to a
    json-based centralized log aggregation platform like ELK.
    n.b., this formatter is very opinionated.
    """
    def format_message(self, record: logging.LogRecord):  # noqa: C901
        s = record.getMessage()
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != '\n':
                s = s + '\n'
            s = s + record.exc_text
        if record.stack_info:
            if s[-1:] != '\n':
                s = s + '\n'
            s = s + self.formatStack(record.stack_info)
        return s

    def iso_time(self, record: logging.LogRecord):
        dt = datetime.utcfromtimestamp(record.created)
        return dt.isoformat('T')

    DEFAULT_KEYS = {
        'name', 'msg', 'args', 'levelname', 'levelno', 'pathname', 'filename',
        'module', 'exc_info', 'exc_text', 'stack_info', 'lineno', 'funcName',
        'created', 'msecs', 'relativeCreated', 'thread', 'threadName',
        'processName', 'process',
    }

    def to_dict(self, record: logging.LogRecord):
        data = {
            'timestamp': self.iso_time(record),
            'message': self.format_message(record),
            'function': record.funcName,
            'path': record.pathname,
            'module': record.module,
            'level': record.levelname,
            'line_number': record.lineno,
        }
        for key, value in record.__dict__.items():
            if key not in JsonFormatter.DEFAULT_KEYS:
                data[key] = value
        return data

    def format(self, record: logging.LogRecord):
        return json.dumps(self.to_dict(record))


class JsonTaskFormatter(JsonFormatter):
    """
    This formatter is useful if you want to ship logs to a
    json-based centralized log aggregation platform like ELK.
    n.b., this formatter is very opinionated.
    """
    def to_dict(self, record: logging.LogRecord):
        data = super(JsonTaskFormatter, self).to_dict(record)
        try:
            from celery._state import get_current_task
            task = get_current_task()
            if task and task.request:
                data['task_id'] = task.request.id
                data['task_name'] = task.name
        except ImportError:
            pass
        return data



import inspect
from time import time
import yaml
from ..utils import flush_print
from ..utils.print_utils import dot_leader


class IntegrationTestException(Exception):
    pass


class IntegrationTest:
    """
    Subclass this class to use for running integration
    tests with the IntegrationTestCommand.  In order
    to use this class, a subclass should define one or
    more instance member functions that begin with the
    word `test` and takes no arguments other than self.
    These functions will be called in alphabetical order
    by the super classes `run_test` function.  Example:

    from djenga.test import IntegrationTest

    class MyIntegrationTest(IntegrationTest):
        def test_addition(self):
            x = 1
            self.assert_equal(x, 1)
            x += 4
            self.assert_equal(x, 5)

    An integration test can also access YAML or JSON test
    data from a file specified by the class variable
    `test_data_file`.  This test data will is made
    available to the sub class through `self.test_data`.
    Example:
        In a file called path/to/my_test_data.yml

        ---
        step_1: 1
        step_2: 5

        In the integration test file:

        class MyIntegrationTest(IntegrationTest):
            test_data_file = 'path/to/my_test_data.yml'

            def test_addition(self):
                x = 1
                self.assert_equal(
                    x, self.test_data['step_1'],
                    'Oh no! failed step 1')
                x += 4
                self.assert_equal(
                    x, self.test_data['step_2'],
                    'Oh no! failed step 2')
    """
    test_data_file = ''

    def __init__(self):
        self.test_data = None
        self.load_test_data()

    def load_test_data(self):
        if not self.test_data_file:
            return
        with open(self.test_data_file, 'r') as f:
            self.test_data = yaml.load(f)

    def assert_true(self, value, message=''):
        if value is not True:
            raise IntegrationTestException(
                message or f'failure: [{value}] was not true')

    def assert_truthy(self, value, message=''):
        if value:
            raise IntegrationTestException(
                message or f'failure: [{value}] was not truthy')

    def assert_false(self, value, message=''):
        if value is not False:
            raise IntegrationTestException(
                message or f'failure: [{value}] was true')

    def assert_falsey(self, value, message=''):
        if value:
            raise IntegrationTestException(
                message or f'failure: [{value}] was not falsey')

    def asset_greater_than(self, left, right, message=''):
        if left <= right:
            raise IntegrationTestException(
                message or f'failure: [{left}] <= [{right}]')

    def asset_greater_than_equal(self, left, right, message=''):
        if left <= right:
            raise IntegrationTestException(
                message or f'failure: [{left}] > [{right}]')

    def assert_less_than(self, left, right, message=''):
        if left >= right:
            raise IntegrationTestException(
                message or f'failure: [{left}] >= [{right}]')

    def assert_less_than_equal(self, left, right, message=''):
        if left > right:
            raise IntegrationTestException(
                message or f'failure: [{left}] > [{right}]')

    def assert_equal(self, left, right, message=''):
        if left != right:
            raise IntegrationTestException(
                message or 'failure: [%s] != [%s]' % (left, right))

    def assert_not_equal(self, left, right, message=''):
        if left == right:
            raise IntegrationTestException(
                message or 'failure: [%s] == [%s]' % (left, right))

    def setup(self):
        pass

    def teardown(self):
        pass

    def run_test(self):
        test_methods = [
            value
            for key, value in inspect.getmembers(self)
            if key.startswith('test') and callable(value)
        ]
        test_methods.sort(key=lambda fn: fn.__name__)
        all_passed = True
        dot_leader('%s.setup', self.__class__.__name__, end='')
        self.setup()
        flush_print('done')
        try:
            for x in test_methods:
                dot_leader(f'{self.__class__.__name__}.{x.__name__}', end='')
                tm_start = time()
                status = 'passed'
                units = 's'
                try:
                    x()
                except IntegrationTestException as ex:
                    status = 'failed (%s)' % (ex,)
                    all_passed = False
                finally:
                    tm_total = time() - tm_start
                if tm_total < 1:
                    tm_total *= 1000
                    units = 'ms'
                flush_print('%s %.1f%s', status, tm_total, units)
        finally:
            dot_leader('%s.teardown', self.__class__.__name__, end='')
            self.teardown()
            flush_print('done')
        return all_passed



from argparse import ArgumentParser
from django.core.management.base import BaseCommand
from ...animal_pairs import animal_pair


class Command(BaseCommand):
    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument(
            '-n', '--count',
            dest='count',
            type=int,
            help='number of animal pairs to output',
            metavar='count',
            default=4,
        )

    def handle(self, count, *args, **options):  # pylint: disable=W0221
        for _ in range(count):
            self.stdout.write(animal_pair())



"""
Encrypts secrets using kms key wrapping.

Usage:
  kms_wrap
    [--region=<region_name>]
    [--profile=<profile name>]
    --key <key_alias_or_id>

Options:
  -r --region=<region_name>    AWS Region Name
  -p --profile=<profile_name>  the name of the profile to use to connect to aws
  -k --key=<key_alias_or_id>   the alias or id of the kms key to use
"""
from argparse import ArgumentParser
import os
import sys
from .kms_wrapped import encrypt


def get_parser():
    parser = ArgumentParser()
    parser.add_argument(
        '-r', '--region',
        dest='region',
        metavar='region_name',
        help='aws region name, e.g., us-east-2',
        default=None,
    )
    parser.add_argument(
        '-p', '--profile',
        dest='profile',
        metavar='profile_name',
        help='the name of the profile to use when connecting to aws',
        default=None,
    )
    parser.add_argument(
        '-k', '--key',
        dest='key',
        metavar='<id or alias>',
        help='the name of the key to use for encryption',
    )
    return parser


def main():
    parser = get_parser()
    args = parser.parse_args()
    # do not print prompt if input is being piped
    if sys.stdin.isatty():
        print('Enter plaintext: ', end='', file=sys.stderr)
        sys.stderr.flush()
    stdin = os.fdopen(sys.stdin.fileno(), 'rb', 0)
    plain_text = stdin.readline()
    plain_text = plain_text.decode('utf-8').rstrip()
    value = encrypt(
        plain_text,
        alias=args.key,
        profile=args.profile,
        region=args.region)
    print(f'{value}')


if __name__ == "__main__":
    main()



from collections import OrderedDict
from celery.backends.redis import RedisBackend


__all__ = [
    'RedisDetailBackend',
    'patch_aliases',
]


class RedisDetailBackend(RedisBackend):
    def _store_result(self, task_id, result, state,
                      traceback=None, request=None, **kwargs):
        """
        :type task_id: str
        :type result: object
        :type status: dict
        :type traceback: object
        :type request: djenga.celery.tasks.StatusContext
        :param kwargs:
        :return:
        """
        if request:
            details = getattr(request, 'details', None)
        else:
            details = kwargs.get('details')
        if isinstance(details, OrderedDict):
            details = [ x.to_json() for x in details.values() ]
        meta = {
            'status': state,
            'result': result,
            'traceback': traceback,
            'children': self.current_task_children(request),
            'details': details,
        }
        self.set(self.get_key_for_task(task_id), self.encode(meta))
        return result


def patch_aliases():
    from celery.app.backends import BACKEND_ALIASES
    BACKEND_ALIASES['redisd'] = 'djenga.celery.backends.RedisDetailBackend'





import importlib
from django.core.management.base import BaseCommand
from django.conf import settings


class Command(BaseCommand):
    @staticmethod
    def queues():
        routes = getattr(settings, 'CELERY_TASK_ROUTES', None)
        routes = routes or {}
        queues = { 'celery' }
        if isinstance(routes, dict):
            for x in routes.values():
                queues.add(x['queue'])
        else:
            if isinstance(routes, (list, tuple)):
                for route_set in routes:
                    for _, x in route_set:
                        queues.add(x['queue'])
        return list(queues)

    def handle(self, *args, **options):
        if settings.DEBUG:
            from celery import platforms
            # print('root: %s' % (C_FORCE_ROOT,))
            from celery.bin import worker
            module = importlib.import_module(settings.DJENGA_CELERY_MODULE)
            app = module.app
            platforms.C_FORCE_ROOT = True
            w = worker.worker(app)
            w.run(loglevel='info', concurrency=1, queues=Command.queues())
        else:
            return "Sorry, I shouldn't be run in production mode (DEBUG=False)"
        return 'Done.'



from collections import OrderedDict, defaultdict
from functools import partial
import logging
import os
import re
from typing import Any
from typing import List
from typing import Dict
from typing import Union
import yaml
from .bunch import Bunch


__all__ = [
    'ConfigBunch',
    'KmsBunch',
    'LazySecret',
    'LazyKmsBunch',
]
logger = logging.getLogger(__name__)


class ConfigBunch(Bunch):
    def __init__(self, *filenames: str, **kwargs):
        """
        initialize with a list of files to be
        read, in order.  nested dicts
        are not supported, but you can
        use dotted key names to achieve the same effect.
        :type filenames: List[str]
        """
        super().__init__()
        for x in filenames:
            try:
                with open(x, 'r') as f:
                    logger.debug('[djenga]  loading settings from [%s]', x)
                    values = self.load_file(f)
                    self.assimilate_values(values)
            except FileNotFoundError:
                logger.debug('[djenga]  file [%s] was not found, skipping.', x)

    def load_file(self, f):
        values: Dict[str, Any] = yaml.load(f, yaml.SafeLoader)
        return values

    def assimilate_values(self, values):
        if not isinstance(values, dict):
            logger.warning('[djenga]  invalid format, skipping')
            return
        for key, value in values.items():
            pieces: List[str] = key.split('.')
            parent = self.create_to_parent(pieces)
            setattr(parent, pieces[-1], value)

    def create_to_parent(self, pieces: List[str]) -> 'ConfigBunch':
        parent: ConfigBunch = self
        for y in pieces[:-1]:
            if not hasattr(parent, y):
                child = ConfigBunch()
                setattr(parent, y, child)
            else:
                child = getattr(parent, y)
            parent = child
        return parent

    def walk_to_parent(self, pieces: List[str]) -> 'ConfigBunch':
        parent: ConfigBunch = self
        for x in pieces[:-1]:
            parent = getattr(parent, x, None)
            if not isinstance(parent, Bunch):
                return None
        return parent

    def get(self, key, default=None):
        pieces = key.split('.')
        parent = self.walk_to_parent(pieces)
        if parent:
            return getattr(parent, pieces[-1], default)
        return default

    def __getitem__(self, key):
        pieces = key.split('.')
        parent = self.walk_to_parent(pieces)
        if parent:
            return getattr(parent, pieces[-1])
        raise KeyError(f'{key} not found in config')

    def __setitem__(self, key, value):
        pieces = key.split('.')
        parent = self.create_to_parent(pieces)
        setattr(parent, pieces[-1], value)

    def setdefault(self, key, value):
        pieces = key.split('.')
        parent = self.create_to_parent(pieces)
        key = pieces[-1]
        if hasattr(parent, key):
            return getattr(parent, key)
        setattr(parent, key, value)
        return value

    def env(self):
        splitter = re.compile(r'[_.]')

        def fn(key, default=None):
            value = os.environ.get(key)
            if not value:
                pieces = splitter.split(key.lower())
                parent: ConfigBunch = self.walk_to_parent(pieces)
                if not isinstance(parent, ConfigBunch):
                    return None
                value = parent.get(pieces[-1], default)
            return value
        return fn

    @staticmethod
    def isdict(values):
        return isinstance(values, (dict, OrderedDict, defaultdict))


class KmsBunch(ConfigBunch):
    def __init__(self, *filenames: str, **kwargs):
        """
        initialize with a list of files to be
        read, in order.  nested dicts
        are not supported, but you can
        use dotted key names to achieve the same effect.
        :type filenames: List[str]
        :type kms_key_id: str
        :type profile: str
        :type region: str
        """
        super().__init__()
        self.profile: str = kwargs.get('profile', None)
        self.region: str = kwargs.get('region', None)
        for x in filenames:
            try:
                with open(x, 'r') as f:
                    logger.debug('[djenga]  loading settings from [%s]', x)
                    values = self.load_file(f)
                    values = self.decrypt(values)
                    self.assimilate_values(values)
            except FileNotFoundError:
                logger.debug('[djenga]  file [%s] was not found, skipping.', x)

    def decrypt(self, values):
        from ..encryption.kms_wrapped import decrypt
        if isinstance(values, (list, tuple)):
            return [ self.decrypt(x) for x in values ]
        if ConfigBunch.isdict(values):
            return {
                key: self.decrypt(value)
                for key, value in values.items()
            }
        try:
            return decrypt(values, region=self.region, profile=self.profile)
        except:  # noqa: E722, pylint: disable=bare-except,
            return values


class LazySecret:
    def __init__(self, value=None, decrypt_fn=None):
        self.decrypted = False
        self.value = value
        self.decrypt = decrypt_fn

    def get(self):
        return self.decrypt(self.value)


class LazyKmsBunch(ConfigBunch):
    def __init__(self, *filenames: str, **kwargs):
        """
        initialize with a list of files to be
        read, in order.  nested dicts
        are not supported, but you can
        use dotted key names to achieve the same effect.
        :type filenames: List[str]
        :type kms_key_id: str
        :type profile: str
        :type region: str
        """
        from ..encryption.kms_wrapped import decrypt
        super().__init__()
        self.profile: str = kwargs.get('profile', None)
        self.region: str = kwargs.get('region', None)
        self.decrypt_fn = partial(
            decrypt, region=self.region, profile=self.profile)
        for x in filenames:
            try:
                with open(x, 'r') as f:
                    logger.debug('[djenga]  loading settings from [%s]', x)
                    values = self.load_file(f)
                    values = self.lazy_wrap(values)
                    self.assimilate_values(values)
            except FileNotFoundError:
                logger.debug('[djenga]  file [%s] was not found, skipping.', x)

    def lazy_wrap(self, values: Union[Dict, List, str]):
        if isinstance(values, (list, tuple)):
            return [ self.lazy_wrap(x) for x in values ]
        if ConfigBunch.isdict(values):
            return {
                key: self.lazy_wrap(value)
                for key, value in values.items()
            }
        if isinstance(values, str) and values.count('|') == 4:
            return LazySecret(values, self.decrypt_fn)
        return values



"""
MySQL database backend for Django.

Requires mysqlclient: https://pypi.org/project/mysqlclient/
"""
from django.db.backends.mysql.base import DatabaseWrapper as _DatabaseWrapper
from .client import DatabaseClient


class DatabaseWrapper(_DatabaseWrapper):
    client_class = DatabaseClient

    def get_connection_params(self):
        from ...core import LazySecret
        kwargs = super().get_connection_params()
        password = kwargs.get('passwd')
        if isinstance(password, LazySecret):
            kwargs['passwd'] = password.get()
        return kwargs



import logging


__all__ = [
    'ZeepHttpsFilter',
    'CeleryRestoringFilter',
]


class ZeepHttpsFilter(logging.Filter):
    def filter(self, record):
        """
        :type record: logging.LogRecord
        :rtype: bool
        """
        result = record.name.startswith('zeep.wsdl.bindings')
        result &= record.levelno == logging.WARNING
        result &= record.msg.startswith('Forcing soap:address location')
        return not result


class CeleryRestoringFilter(logging.Filter):
    def filter(self, record):
        """
        :type record: logging.LogRecord
        :rtype: bool
        """
        result = record.name.startswith('celery.redirected')
        result &= record.msg.startswith('Restoring')
        result &= record.msg.endswith('unacknowledged message(s)')
        return not result



from .helpers import _as_bytes
from .helpers import from_b64_str
from .helpers import _get_client
from .helpers import _prefix_alias
from .gcm import encrypt_bytes as gcm_encrypt
from .gcm import decrypt_bytes as gcm_decrypt


def encrypt_bytes(
        plain_text: bytes
        , alias: str
        , region: str = None
        , profile: str = None) -> str:
    client = _get_client(region, profile)
    alias = _prefix_alias(alias)
    response = client.generate_data_key(KeyId=alias, KeySpec='AES_256')
    data_key = response['Plaintext']
    header = response['CiphertextBlob']
    value = gcm_encrypt(plain_text, data_key, auth_header=header)
    return value


def decrypt_bytes(
        packed_value: str
        , region: str = None
        , profile: str = None) -> bytes:
    pieces = packed_value.split('|', 1)
    wrapped_data_key = pieces[0]
    wrapped_data_key = from_b64_str(wrapped_data_key)
    client = _get_client(region, profile)
    response = client.decrypt(CiphertextBlob=wrapped_data_key)
    data_key = response['Plaintext']
    plain_text = gcm_decrypt(packed_value, data_key)
    return plain_text


def encrypt(plain_text, alias, region: str = None, profile: str = None) -> str:
    plain_text = _as_bytes(plain_text)
    data = encrypt_bytes(plain_text, alias, region, profile)
    return data


def decrypt(packed_value: str, region: str = None, profile: str = None):
    data = decrypt_bytes(packed_value, region, profile)
    return data.decode('utf-8')



# encoding: utf-8
from .util import ColorFormatter
from .util import BriefColorFormatter
from .util import IgnoreCaching
from .util import IgnoreDjangoInternals
from .util import IgnorePisaPdf
from .util import IgnoreRequests
from .util import WorldWritableRotatingFileHandler
from .filters import *




from .memory import *  # noqa: F401, F403
from .timing import start_timer
from .timing import end_timer

__all__ = [
    'start_timer',
    'end_timer',
    'get_memory_usage',
    'get_peak_memory_usage',
]





from base64 import b64encode
from base64 import b64decode
from threading import local
import boto3
import six


__all__ = [
    '_as_bytes',
    'b64_str',
    'from_b64_str',
    '_get_client',
    '_prefix_alias',
]


thread_local = local()
thread_local.sessions = {}


def _as_bytes(value):
    if isinstance(value, six.string_types):
        value = value.encode('utf-8')
    return value


def b64_str(value: bytes):
    return b64encode(value).decode('utf-8')


def from_b64_str(value: str):
    value = value.encode('utf-8')
    return b64decode(value)


def _get_client(region: str = None, profile: str = None):
    key = f'{region}-{profile}'
    client = thread_local.sessions.get(key)
    if not client:
        session = boto3.Session(region_name=region, profile_name=profile)
        client = session.client('kms')
        thread_local.sessions[key] = client
    return client


def _prefix_alias(alias: str):
    if not alias.startswith('alias/'):
        alias = f'alias/{alias}'
    return alias





import csv


class Utf8Recoder:
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8
    """
    def __init__(self, f):
        super(Utf8Recoder, self).__init__()
        self.reader = f

    def __iter__(self):
        return self.reader

    def next(self):
        return self.reader.next().encode('utf-8')


UnicodeCsvReader = csv.reader


def read_unicode_csv(filename, encoding='utf-8'):
    with open(filename, 'r', encoding=encoding) as f:
        reader = UnicodeCsvReader(f)
        for x in reader:
            yield x




import sys
import re
from functools import partial


def flush_print(st, *args, **kwargs):
    end = kwargs.pop('end', '\n')
    if args:
        st = st % args
    print(st, end=end)
    if sys.stdout.isatty():
        sys.stdout.flush()


def cprint(color_fn, st, *args):
    if args:
        st = st % args
    print(color_fn(st), end='')
    sys.stdout.flush()


def dot_lead(st, *args, **kwargs):
    width = kwargs.pop('width', 60)
    if args:
        st = st % args
    dots = '.' * (width - len(st))
    return '%s%s' % (st, dots)


def dot_leader(st, *args, **kwargs):
    width = kwargs.pop('width', 60)
    if args:
        st = st % args
    dots = '.' * (width - len(st))
    flush_print('%s%s', st, dots, **kwargs)


COLORS = (
    'black', 'red', 'green', 'yellow',
    'blue', 'magenta', 'cyan', 'white')
STYLES = (
    'bold', 'faint', 'italic', 'underline',
    'blink', 'blink2', 'negative',
    'concealed', 'crossed')


def _ansi_color(x, offset, ansi_code):
    if x:
        if x in COLORS:
            return [ f'{offset + COLORS.index(x)}' ]
        if isinstance(x, int) and 0 <= x <= 255:
            return [ f'{ansi_code};5;{x}' ]
        raise Exception('Invalid color [{x}]')
    return []


def _style_codes(style):
    codes = []
    if style:
        for st in style.split('+'):
            if st in STYLES:
                codes.append(f'{(1 + STYLES.index(st))}')
            else:
                raise Exception('Invalid style "{st}"')
    return codes


def color(s, fg=None, bg=None, style=None):
    sgr = _ansi_color(fg, 30, 38) + _ansi_color(bg, 40, 48)
    sgr += _style_codes(style)
    if sgr:
        return f"\x1b[{';'.join(sgr)}m{s}\x1b[0m"
    return s


def strip_color(s):
    return re.sub(r'\x1b\[.+?m', '', s)


# Foreground shortcuts
black = partial(color, fg='black')
red = partial(color, fg='red')
green = partial(color, fg='green')
yellow = partial(color, fg='yellow')
blue = partial(color, fg='blue')
magenta = partial(color, fg='magenta')
cyan = partial(color, fg='cyan')
white = partial(color, fg='white')

# Style shortcuts
bold = partial(color, style='bold')
faint = partial(color, style='faint')
italic = partial(color, style='italic')
underline = partial(color, style='underline')
blink = partial(color, style='blink')
blink2 = partial(color, style='blink2')
negative = partial(color, style='negative')
concealed = partial(color, style='concealed')
crossed = partial(color, style='crossed')



from .commandlogginglevels import VerbosityLevels
from .configurationmixin import ConfigurationMixin
from .csvmixin import CsvMixin
from .inputcsvmixin import InputCsvMixin
from .statuscommand import StatusCommand
from .run_integration_tests import Command as IntegrationTestCommand

__all__ = [
    'VerbosityLevels',
    'ConfigurationMixin',
    'CsvMixin',
    'InputCsvMixin',
    'StatusCommand',
    'IntegrationTestCommand',
]
